/*********************************************/
/* Exemplo de uso de macros e funes para   */
/* tratamento de excees de ponto flutuante */
/*                                           */
/* Funes demonstradas:                     */
/*       * feholdexcept()                    */
/*       * feupdateenv()                     */
/*       * feraiseexcept()                   */
/*       * fetestexcept()                    */
/*********************************************/

#include <stdio.h>
#include <fenv.h>

void ImprimeFlags(const char* textoInicial, int flags)
{
   if (flags & FE_INEXACT)
      printf("\n%s\n\tOcorreu um resultado inexato\n",
             textoInicial);

   if (flags & FE_UNDERFLOW)
      printf("\n%s\n\tOcorreu underflow\n",
             textoInicial);

   if (flags & FE_OVERFLOW)
      printf("\n%s\n\tOcorreu overflow\n",
             textoInicial);

   if (flags & FE_DIVBYZERO)
      printf("\n%s\n\tOcorreu uma divisao por zero\n",
             textoInicial);

   if (flags & FE_INVALID)
      printf("\n%s\n\tOcorreu uma operacao invalida\n",
             textoInicial);

   if (!(flags & FE_ALL_EXCEPT))
      printf("\n%s\n\tNao houve ocorrencia de excecao\n",
             textoInicial);
}

double OperacaoDePF(double x)
{
   fenv_t  ambientePF;
   int     excecao;

      /* O trecho a seguir demonstra que, como esta */
      /* funo pode ter sido chamada depois de     */
      /* alguma flag ter sido ligada, se ocorrer    */
      /* alguma exceo aqui, no ser possvel     */
      /* saber se esta exceo surgiu nesta funo  */
      /* ou se a mesma j havia ocorrido antes de   */
      /* esta funo ser chamada.                   */
   excecao = fetestexcept(FE_ALL_EXCEPT);
   ImprimeFlags( "Na entrada da funcao OperacaoDePF:",
                 excecao );

      /* Aqui comea a situao real. Primeiro,   */
      /* salvam-se os sinalizadores e desligam-se */
      /* os mesmos usando a funo feholdexcept() */
   feholdexcept(&ambientePF);

      /* Agora so executadas operaes de ponto */
      /* flutuante que podem dar origem a alguma */
      /* exceo. Aqui, isto  simulado usando   */
      /* a funo feraiseexcept().               */

   // ... (operaes que podem causar exceo)
   feraiseexcept(FE_INEXACT);

      /* Numa situao real, as instrues a seguir   */
      /* seriam substitudas por instrues que testa */
      /* se ocorreram execees nas operaes acima.  */
   excecao = fetestexcept(FE_ALL_EXCEPT);
   ImprimeFlags("Na funcao OperacaoDePF, depois da "
                "ocorrencia de uma excecao:", excecao);

      /* Agora as excees de ponto flutuante que     */
      /* possam ter ocorrido seriam tratadas a seguir */
   // ... (tratamento de evetuais excees)

      /* Restaura a configurao de ambiente    */
      /* guardada acima e liga os sinalizadores */
      /* de excees que ocorreram nesta funo */
   feupdateenv(&ambientePF);

   excecao = fetestexcept(FE_ALL_EXCEPT);
   ImprimeFlags( "Na funcao OperacaoDePF, antes "
                 "de retornar:", excecao );

   return x;
}

int main(void)
{
  int       excecao;

  		/* Suponha que no trecho de programa a seguir */
  		/* possa ocorrer uma situao de exceo.     */
  		/* Aqui, ns simulamos isto usando a funo   */
  		/* feraiseexcept().                           */

   // ... (operaes que podem causar exceo)
   feraiseexcept(FE_DIVBYZERO);

   excecao = fetestexcept(FE_ALL_EXCEPT);
   ImprimeFlags( "Na funcao main antes de chamar "
                 "OperacaoDePF:", excecao );

      /* Agora, chama a funo OperacaoDePF()  */
      /* que tambm pode provocar uma exceo. */
   OperacaoDePF(2.5);

   excecao = fetestexcept(FE_ALL_EXCEPT);
   ImprimeFlags( "Na funcao main depois de chamar "
                 "OperacaoDePF:", excecao );

   return 0;
}

/*

Resultado do programa no Windows XP:

Na funcao main antes de chamar OperacaoDePF:
        Ocorreu uma divisao por zero

Na entrada da funcao OperacaoDePF:
        Ocorreu uma divisao por zero

Na funcao OperacaoDePF, depois da ocorrencia de uma excecao:
        Ocorreu um resultado inexato

Na funcao OperacaoDePF, antes de retornar:
        Ocorreu um resultado inexato

Na funcao OperacaoDePF, antes de retornar:
        Ocorreu uma divisao por zero

Na funcao main depois de chamar OperacaoDePF:
        Ocorreu um resultado inexato

Na funcao main depois de chamar OperacaoDePF:
        Ocorreu uma divisao por zero

*/
